# ----------------------------------------------------------------------
#  COMPONENT: DrawingEntry - widget for entering numeric values
#
#  This widget represents a <number> entry on a control panel.
#  It is used to enter numeric values.
# ======================================================================
#  AUTHOR:  Michael McLennan, Purdue University
#  Copyright (c) 2004-2005  Purdue Research Foundation
#
#  See the file "license.terms" for information on usage and
#  redistribution of this file, and for a DISCLAIMER OF ALL WARRANTIES.
# ======================================================================
package require Itk

itcl::class Rappture::DrawingEntry {
    inherit itk::Widget
    itk_option define -state state State "normal"

    private variable _owner
    private variable _path
    private variable _cname2id
    private variable _xMin 0
    private variable _yMin 0
    private variable _xScale 0
    private variable _yScale 0
    private variable _xOffset 0
    private variable _yOffset 0
    private variable _canvasWidth 0
    private variable _canvasHeight 0
    private variable _xAspect 0
    private variable _yAspect 0
    constructor {owner path args} { # defined below }

    public method value {args}

    public method label {}
    public method tooltip {}

    private method AdjustDrawingArea { xAspect yAspect } 
    private method Highlight { tag } 
    private method ParseDescription {}
    private method ParseGrid { cpath cname }
    private method ParseHotspot { cpath cname }
    private method ParseLine { cpath cname }
    private method ParseOval { cpath cname }
    private method ParsePicture { cpath cname }
    private method ParsePolygon { cpath cname }
    private method ParseRectangle { cpath cname }
    private method ParseText { cpath cname }
    private method Redraw {}
    private method ScreenCoords { coords } 
    private method ScreenX { x } 
    private method ScreenY { y } 
    private method Unhighlight { tag } 
}

itk::usual DrawingEntry {
    keep -cursor -font
    keep -foreground -background
    keep -textbackground
    keep -selectbackground -selectforeground -selectborderwidth
}

# ----------------------------------------------------------------------
# CONSTRUCTOR
# ----------------------------------------------------------------------
itcl::body Rappture::DrawingEntry::constructor {owner path args} {
    if {[catch {$owner isa Rappture::ControlOwner} valid] != 0 || !$valid} {
        error "bad object \"$owner\": should be Rappture::ControlOwner"
    }
    set _path $path
    set _owner $owner
    #
    # Display the current drawing.
    #
    itk_component add drawing {
	canvas $itk_interior.canvas -background white -relief sunken -bd 1
    } {
	ignore -background
    }
    pack $itk_component(drawing) -expand yes -fill both
    ParseDescription
    eval itk_initialize $args
}

# ----------------------------------------------------------------------
# USAGE: value ?-check? ?<newval>?
#
# Clients use this to query/set the value for this widget.  With
# no args, it returns the current value for the widget.  If the
# <newval> is specified, it sets the value of the widget and
# sends a <<Value>> event.  If the -check flag is included, the
# new value is not actually applied, but just checked for correctness.
# ----------------------------------------------------------------------
itcl::body Rappture::DrawingEntry::value {args} {
puts "value $args"
    set onlycheck 0
    set i [lsearch -exact $args -check]
    if {$i >= 0} {
        set onlycheck 1
        set args [lreplace $args $i $i]
    }

    if {[llength $args] == 1} {
        if {$onlycheck} {
            # someday we may add validation...
            return
        }
        set xmlobj [lindex $args 0]
        $itk_component(drawing) value $xmlobj
        return $xmlobj

    } elseif {[llength $args] != 0} {
        error "wrong # args: should be \"value ?-check? ?newval?\""
    }

    #
    # Query the value and return.
    #
    return [$itk_component(drawing) value]
}

# ----------------------------------------------------------------------
# USAGE: label
#
# Clients use this to query the label associated with this widget.
# Reaches into the XML and pulls out the appropriate label string.
# ----------------------------------------------------------------------
itcl::body Rappture::DrawingEntry::label {} {
return ""
    set label [$_owner xml get $_path.about.label]
    if {"" == $label} {
        set label "Drawing"
    }
    return $label
}

# ----------------------------------------------------------------------
# USAGE: tooltip
#
# Clients use this to query the tooltip associated with this widget.
# Reaches into the XML and pulls out the appropriate description
# string.  Returns the string that should be used with the
# Rappture::Tooltip facility.
# ----------------------------------------------------------------------
itcl::body Rappture::DrawingEntry::tooltip {} {
return ""
    set str [$_owner xml get $_path.about.description]
    return [string trim $str]
}

# ----------------------------------------------------------------------
# CONFIGURATION OPTION: -state
# ----------------------------------------------------------------------
itcl::configbody Rappture::DrawingEntry::state {
    set valid {normal disabled}
    if {[lsearch -exact $valid $itk_option(-state)] < 0} {
        error "bad value \"$itk_option(-state)\": should be [join $valid {, }]"
    }
}

itcl::body Rappture::DrawingEntry::Redraw {args} {
    # Remove exists canvas items and hints
    $itk_component(drawing) delete all
    # Recompute the size of the canvas/drawing area
    set _canvasWidth [winfo width $itk_component(drawing)] 
    if { $w < 2 } {
	set _canvasWidth [winfo reqwidth $itk_component(drawing)]
    }
    set _canvasHeight [winfo height $itk_component(drawing)]
    if { $_canvasHeight < 2 } {
	set _canvasHeight [winfo reqheight $itk_component(drawing)]
    }
    set _drawingWidth $_canvasWidth
    set _drawingHeight $_canvasHeight
    set _xOffset 0
    set _yOffset 0
    array set _hints {
	set background white
    }
    ParseDecription
}

itcl::body Rappture::DrawingEntry::ParseLine { cpath cname } {
    array set attr2option {
	"linewidth"	"-width"
	"arrow"		"-arrow"
	"dash"		"-dash"
	"color"		"-fill"
    }
    # Set default options first and then let tool.xml override them.
    array set options {
	-arrow		none
	-width		0
	-fill		black
	-dash		""
    }
    # Coords
    set coords {}
    set coords [$_owner xml get $cpath.coords]
    set coords [string trim $coords]
    if { $coords == "" } {
	set coords "0 0"
    } else {
	set coords [ScreenCoords $coords]
    }
    puts stderr "ParseLineDescrption description owner=$_owner cpath=$cpath coords=$coords"
    set list {}
    foreach attr [$_owner xml children $cpath] {
	if { [info exists attr2option($attr)] } {
	    set option $attr2option($attr)
	    set value [$_owner xml get $cpath.$attr]
	    set options($option) $value
	}
    }
    set options(-tags) $cname
    puts stderr "$itk_component(drawing) create line $coords"
    set id [eval $itk_component(drawing) create line $coords]
    set _cname2id($cname) $id
    eval $itk_component(drawing) itemconfigure $id [array get options]
}

#
# ParseGrid -- 
#
itcl::body Rappture::DrawingEntry::ParseGrid { cpath cname } {
    puts stderr "ParseGrid description owner=$_owner cpath=$cpath"
    foreach attr [$_owner xml children $cpath] {
	puts stderr attr=$attr
    }
    array set attr2option {
	"linewidth"	"-width"
	"arrow"		"-arrow"
	"dash"		"-dash"
	"color"		"-fill"
    }
    # Set default options first and then let tool.xml override them.
    array set options {
	-arrow		none
	-width		0
	-fill		black
	-dash		""
    }
    # Coords
    set xcoords [$_owner xml get $cpath.xcoords]
    set xcoords [string trim $xcoords]
    set ycoords [$_owner xml get $cpath.ycoords]
    set ycoords [string trim $ycoords]
    if { $ycoords == "" } {
	set ycoords "0 1"
	set ymax 1 
	set ymin 0
    } else {
	set list {}
	set ymax -10000 
	set ymin 10000
	foreach c $ycoords {
	    set y [ScreenY $c]
	    if { $y > $ymax } {
		set ymax $y
	    } 
	    if { $y < $ymin } {
		set ymin $y
	    }
	    lappend list $y
	}
	set ycoords $list
    }
    if { $xcoords == "" } {
	set xcoords "0 1"
	set xmax 1 
	set xmin 0
    } else {
	set list {}
	set xmax -10000 
	set xmin 10000
	foreach c $xcoords {
	    set x [ScreenX $c]
	    if { $x > $xmax } {
		set xmax $x
	    } 
	    if { $x < $xmin } {
		set xmin $x
	    }
	    lappend list $x
	}
	set xcoords $list
    }
    puts stderr "ParseGrid description owner=$_owner cpath=$cpath xcoords=$xcoords ycoords=$ycoords"
    set list {}
    foreach attr [$_owner xml children $cpath] {
	if { [info exists attr2option($attr)] } {
	    set option $attr2option($attr)
	    set value [$_owner xml get $cpath.$attr]
	    set options($option) $value
	}
    }
    set options(-tags) $cname
    foreach y $ycoords {
	lappend ids \
	    [eval $itk_component(drawing) create line $xmin $y $xmax $y \
		 [array get options]]
    }
    foreach x $xcoords {
	lappend ids \
	    [eval $itk_component(drawing) create line $x $ymin $x $ymax \
		 [array get options]]
    }
    set _cname2id($cname) $ids
}

itcl::body Rappture::DrawingEntry::ParseText { cpath cname } {
    array set attr2option {
	"font"		"-font"
	"default"	"-text"
	"color"		"-fill"
	"text"		"-text"
	"anchor"	"-anchor"
    }
    puts stderr "ParseText description owner=$_owner cpath=$cpath"

    # Set default options first and then let tool.xml override them.
    array set options {
	-font {Arial 12}
	-text {}
	-fill black
	-anchor c
    }
    foreach attr [$_owner xml children $cpath] {
	if { [info exists attr2option($attr)] } {
	    set option $attr2option($attr)
	    set value [$_owner xml get $cpath.$attr]
	    set options($option) $value
	}
    }
    # Coords
    set coords {}
    set coords [$_owner xml get $cpath.coords]
    set coords [string trim $coords]
    if { $coords == "" } {
	set coords "0 0"
    } else {
	set coords [ScreenCoords $coords]
    }
    set options(-tags) $cname
    puts stderr "$itk_component(drawing) create text $coords"
    set id [eval $itk_component(drawing) create text $coords]
    set _cname2id($cname) $id
    puts stderr "$itk_component(drawing) itemconfigure $id [array get options]"
    eval $itk_component(drawing) itemconfigure $id [array get options]
}

itcl::body Rappture::DrawingEntry::ParseRectangle { cpath cname } {
    array set attr2option {
	"outline"	"-outline"
	"fill"		"-fill"
	"linewidth"	"-linewidth"
    }
    puts stderr "ParseRectangle description owner=$_owner cpath=$cpath"

    # Set default options first and then let tool.xml override them.
    array set options {
	-fill blue
	-linewidth 1 
	-outline black
    }
    foreach attr [$_owner xml children $cpath] {
	if { [info exists attr2option($attr)] } {
	    set option $attr2option($attr)
	    set value [$_owner xml get $cpath.$attr]
	    set options($option) $value
	}
    }
    # Coordinates
    set coords {}
    set coords [$_owner xml get $cpath.coords]
    set coords [string trim $coords]
    if { $coords == "" } {
	set coords "0 0 1 1"
    }
    foreach { x y } [ScreenCoords $coords] break
    set lx $x
    set ly $y

    puts stderr "$itk_component(drawing) itemconfigure $id [array get options]"
    eval $itk_component(drawing) itemconfigure $id [array get options]
    set bbox [$itk_component(drawing) bbox $id]
    puts stderr "cpath=$cpath bbox=$bbox"
    set sid [eval $itk_component(drawing) create rectangle $bbox]
    $itk_component(drawing) itemconfigure $sid -fill "" -outline "" \
	-tag $cname-bg
    set sid [eval $itk_component(drawing) create rectangle $bbox]
    $itk_component(drawing) itemconfigure $sid -fill "" -outline "" \
	-tag $cname-sensor
    $itk_component(drawing) bind $cname-sensor <Enter> \
	[itcl::code $this Highlight $cname]
    $itk_component(drawing) bind $cname-sensor <Leave> \
	[itcl::code $this Unhighlight $cname]
    $itk_component(drawing) lower $cname-bg
    $itk_component(drawing) raise $cname-sensor
}

itcl::body Rappture::DrawingEntry::ScreenX { x } {
    set norm [expr ($x - $_xMin) * $_xScale]
    set x [expr int($norm * $_drawingWidth) + $_xOffset]
    return $x
}

itcl::body Rappture::DrawingEntry::ScreenY { y } {
    set norm [expr ($y - $_yMin) * $_yScale]
    set y [expr int($norm * $_drawingHeight) + $_yOffset]
    return $y
}

itcl::body Rappture::DrawingEntry::ScreenCoords { coords } {
    set list {}
    foreach {x y} $coords {
	lappend list [ScreenX $x] [ScreenY $y]
    }
    return $list
}

itcl::body Rappture::DrawingEntry::AdjustDrawingArea { xAspect yAspect } {
    set _drawingWidth $_canvasWidth 
    set _drawingHeight $_canvasHeight
    if { $xAspect <= 0 || $yAspect <= 0 } {
	return
    }
    set current [expr double($_canvasWidth) / double($_canvasHeight)]
    set wanted [expr double($xAspect) / double($yAspect)]
    if { $current > $wanted } {
	set sw [ expr int($_canvasWidth * $wanted)]
	if { $sw < 1 } {
	    set sw 1
	}
	set _xOffset [expr $_canvasWidth - $sw]
	set _drawingWidth $sw
    } else {
	set sh [expr int($_canvaseHeight / $wanted)]
	if { $sh < 1 }  {
	    set sh 1
	}
	set _xOffset [expr $_canvasHeight - $sh]
	set _drawingHeight $sh
    }
}

itcl::body Rappture::DrawingEntry::ParseDescription {} {
    puts stderr "ParseDescrption description owner=$_owner path=$_path"
    ParseBackground
    ParseSubstitutions
    foreach cname [$_owner xml children $_path.components] {
	switch -glob -- $cname {
	    "line*" {
		ParseLine $_path.components.$cname
	    }
	    "grid*" {
		ParseGrid $_path.components.$cname
	    }
	    "text*" {
		ParseText $_path.components.$cname
	    }
	    "picture*" {
		ParsePicture $_path.components.$cname
	    }
	    "rectangle*" {
		ParseRectangle $_path.components.$cname
	    }
	    "oval*" {
		ParseOval $_path.components.$cname
	    }
	    "polygon*" {
		ParsePolygon $_path.components.$cname
	    }
	    "hotspot*" {
		ParseHotspot $_path.components.$cname
	    }
	}
    }
}



itcl::body Rappture::DrawingEntry::Highlight { tag } {
    $itk_component(drawing) itemconfigure $tag-bg -outline black \
	-fill lightblue
}

itcl::body Rappture::DrawingEntry::Unhighlight { tag } {
    $itk_component(drawing) itemconfigure $tag-bg -outline "" \
	-fill ""
}

#
#      <background>
#       <!-- background color of the drawing canvas (default white) -->
#       <color>black</color>
#       <!-- coordinate system:  x0 y0 ?at screenx screeny? x1 y1 
#				?at screenx screeny?
#            The screenx/screeny values are optional, so you can also say
# 	   something like "-.1 0 1.1 1" as you had in your example.
# 	   This lets you put the origin at a specific point on screen,
# 	   and also define the directions of the axes.  We still compute
# 	   the overall bounding box.  In the example below, the bounding
# 	   box goes from -1,1 in the upper-left corner to 1,-1 in the
# 	   lower right.
#       -->
#       <coordinates>0 0 at 50% 50% 1 1 at 100% 100%</coordinates>

#       <!-- aspect ratio:  scales coordinate system so that pixels may not
#            be square.  A coordinate system like the one above implies a
# 	   square drawing area, since x and y both go from -1 to 1.  But
# 	   if you set the aspect to 2:1, you'll get something twice as
# 	   wide as it is tall.  This effectively says that x goes from
# 	   -1 to 1 in a certain distance, but y goes from -1 to 1 in half
# 	   that screen distance.  Default is whatever aspect is defined
# 	   by the coordinates.  If x goes 0-3 and y goes 0-1, then the
# 	   drawing (without any other aspect ratio) would be 3x wide and
# 	   1x tall.  The aspect ratio could be used to force it to be
# 	   square instead by setting "1 1" instead.  In that case, x goes
# 	   0-3 over the width, and y goes 0-1 over the same screen distance
# 	   along the height.
#       -->
#       <aspect>2 1</aspect>
#     </background>
#

itcl::body Rappture::DrawingEntry::ParseCoordinates {} {
    set len [llength $values]
    if { $len == 4 } {
	if { [scan $values "%g %g %g %g" x1 y1 x2 y2] != 4 } {
	    error "bad coordinates specification \"$values\""
	}
	set _xScale [expr 1.0 / ($x2 - $x1)]
	set _yScale [expr 1.0 / ($y2 - $y2)]
	set _xMin $x1
	set _yMin $y1
    } elseif { $len == 10 } {
	if { [scan $values "%g %g %s %d%% %d%% %g %g %s %d%% %d%%" \
		  sx1 sy1 at1 x1 y1 sx2 sy2 at2 x2 y2] != 10 } {
	    error "bad coordinates specification \"$values\""
	}
	if { $at1 != "at" || $at2 != "at" } {
	    error "bad coordinates specification \"$values\""
	}	    
	set x1 [expr $x1 / 100.0]
	set x2 [expr $x2 / 100.0]
	set y1 [expr $y1 / 100.0]
	set y2 [expr $y2 / 100.0]
	set _xScale [expr ($sx2 - $sx1) / ($x2 - $x1)]
	set _yScale [expr ($sy2 - $sy2) / ($y2 - $y2)]
	set _xMin $x1
	set _yMin $y1
    }
}

itcl::body Rappture::DrawingEntry::ParseBackground {} {
    foreach elem [$_owner xml children $_path.background] {
	switch -- $elem {
	    #  Background color of the drawing canvas (default white)
	    "color" {
		set value [$_owner xml get $_path.background.$elem]
		set _hints(background) $value
	    }
	    "aspect" {
		set value [$_owner xml get $_path.background.$elem]
		foreach { xAspect yAspect } $value break
		AdjustDrawingArea $xAspect $yAspect
	    }
	    "coordinates" {
		set value [$_owner xml get $_path.background.$elem]
		ParseCoordinates $value
	    }
	}
    }
}

itcl::body Rappture::DrawingEntry::ParseSubstitutions {} {
   foreach var [$_owner xml children $_path.substitutions] {
       if { ![string match "variable*" $var] } {
	   continue
       }
       set varPath $_path.substitutions.$var
       foreach elem [$_owner xml children $varPath] {
	   set map ""
	   set name ""
	   set path ""
	   switch -- $elem {
	    #  Background color of the drawing canvas (default white)
	    "name" {
		set name [$_owner xml get $varPath.$elem]
	    }
	    "path" {
		set path [$_owner xml get $varPath.$elem]
	    }
	    "map" {
		set from [$_owner xml get $varPath.$elem.from]
		set to [$_owner xml get $varPath.$elem.to]
		if { $from == "" || $to == "" } {
		    puts stderr "empty translation in map table \"$varPath\""
		}
		lappend map $from $to
	    }
	    if { $name == "" } {
		puts stderr \
		    "no name defined for substituion variable \"$varPath\""
		continue
	    }
	    if { [info exists _name2path($name)] } {
		puts stderr \
		    "substitution variable \"$name\" already defined"
		continue
	    }		
	    set _name2path($name) $path
	    if { $path == "" } {
		puts stderr \
		    "no path defined for substituion variable \"$varPath\""
		continue
	    }
	    set _name2map($name) $map
	}
    }
}
